# engNTRU OpenSSL ENGINE

This project aims at developing an OpenSSL ENGINE to provide a
batch implementation for KEM cryptographic primitives, particularly the
post-quantum algorithm sntrup761.

The project currently aims at compability with OpenSSL releases in the
`1.1.1` LTS branch.

## Limitations

-   The project currently supports two back-end providers:
    -   [`serial_lib`](providers/serial_lib): key generation is not done in
        batches, but a new fresh keypair is generated on-demand for each
        request.
    -   [`batch_lib`](providers/batch_lib): key generation is done in batches
        that are handled as process local, i.e., each thread of an application
        linking against OpenSSL and loading this ENGINE will get a fresh keypair
        from a precomputed batch, or on-demand generate a fresh batch if the
        current batch does not contain fresh keypairs anymore.
        Each process will handle its own batches indipendently.
        **CAVEAT**: if a process loads this ENGINE, generates a batch and then
        forks, it will likely result in potential keypair reuse.

-   The project currently only uses batch generation in the context of key
    generation;

-   The project currently supports ephemeral keys and does not allow to
    serialize keys as files;

## Table of Contents

<!-- toc -->

- [Project structure](#project-structure)
- [Installation](#installation)
  * [Binary distributions of OpenSSL](#binary-distributions-of-openssl)
  * [Installing prerequisites from source](#installing-prerequisites-from-source)
    + [OpenSSL](#openssl)
  * [Install engNTRU from sources](#install-engntru-from-sources)
  * [Uninstall](#uninstall)
- [Usage](#usage)
  * [List algorithms](#list-algorithms)
  * [TLS test](#tls-test)
    + [Generate an RSA cert for the server](#generate-an-rsa-cert-for-the-server)
    + [Launch the server](#launch-the-server)
    + [Launch the client](#launch-the-client)
  * [ENV variables](#env-variables)
    + [`ENGNTRU_DEBUG`](#engntru_debug)
    + [OpenSSL ENV variables](#openssl-env-variables)
- [License](#license)
- [Acknowledgments](#acknowledgments)

<!-- tocstop -->

## Project structure

The source code of the project is organized hierarchically.

```
engntru
├── debug
├── meths
│   └── internal
│       └── CMakeLists.txt
├── ossl
├── providers
│   ├── api
│   ├── _dummy
│   │   └── CMakeLists.txt
│   ├── batch_lib
│   │   └── CMakeLists.txt
│   └── serial_lib
│       └── CMakeLists.txt
├── tests
│   └── CMakeLists.txt
└── engntru.c
```

-   `engntru.c` contains the main entry point for loading the `ENGINE`;
-   `meths` contains the implementation of the OpenSSL method structures
    defining the implemented cryptosystems, i.e.,
    `EVP_PKEY_METHOD`/`EVP_PKEY_ASN1_METHOD`;
-   `ossl` contains code to integrate error codes, messages, NIDs, and
    OIDs in the OpenSSL abstractions;
-   `providers` contains the code to map the primitives referenced in
    the `meths` structures to the actual cryptographic implementation
    provider:
    *   `api` describes the API that a valid provider module needs to
        implement;
    *   `_dummy` includes boilerplate code for additional functions
        (e.g. an empty `engntru_implementation_init()` that can be used
        when the backend provider does not require any initialization
        before being used);
    *   `serial-lib`, `batch_lib` map the cryptographic functionality to the
        corresponding backend implementation;
-   `tests` contains code used to automate testing of the `ENGINE`;
-   `debug` contains definitions used to implement the debug messaging
    system;

## Installation

To build `engNTRU` from source you will need:
-   `git` to clone the latest source version from this repository and
    other dependencies you plan to build from source;
-   `cmake`, `pkg-config`, `ruby`, `make`, `gcc`/`clang` and the
    required development headers specific for your system, to ensure a
    working build system.

In Debian-like distributions the following should suffice:

```
apt-get install git pkg-config cmake ruby build-essential
```

Other flavours of UNIX will use a different package manager (replacing
`apt-get install` with something similar) and use slightly different
package names.

### Binary distributions of OpenSSL

**Note**: verify that your OS vendor provides OpenSSL >= `1.1.1`

**Note**: currently using KEM in TLS requires a patched version of OpenSSL's
`libssl`.


To use OpenSSL as provided by your Linux distribution, you
need to make sure the development headers are also installed.

In Debian/Ubuntu this means to install the corresponding `*-dev`
packages:

```
apt-get install libssl-dev
```

**Note**: the above step is not required if installing OpenSSL from
source.

Other flavours of UNIX will use a different package manager (replacing
`apt-get install` with something similar) and use slightly different
package names.

### Installing prerequisites from source

#### OpenSSL

**TODO**: fix branch name and git link to match the custom version of OpenSSL
with patched `libssl` for KEM

```
git clone --single-branch --branch OpenSSL_1_1_1-stable https://github.com/openssl/openssl.git openssl-111
cd openssl-111/
export OPENSSL_PREFIX=/usr/local/ssl
./config -d shared --prefix=$OPENSSL_PREFIX --openssldir=$OPENSSL_PREFIX -Wl,-rpath=$OPENSSL_PREFIX/lib
make -j8
make test
# either make install_sw or checkinstall
sudo checkinstall --strip=no --stripso=no --pkgname=openssl-111-debug --provides=openssl-111-debug --pkgversion=1.1.1 --default make install_sw
alias openssl=$OPENSSL_PREFIX/bin/openssl
```

**TODO**: add steps required to install `libsntrup761_batch` from sources

### Install engNTRU from sources

```
git clone <TODO>/engntru
cd engntru
mkdir build
cd build
cmake -DCMAKE_BUILD_TYPE=Debug -DCMAKE_PREFIX_PATH="${OPENSSL_PREFIX};${LIBSNTRUP761BATCH_PREFIX}" ..
make
make test
#ctest --output-on-failure
#ctest --verbose
#
# either make install or checkinstall
sudo checkinstall --strip=no --stripso=no --pkgname=engntru-debug --provides=engntru-debug --default
```



### Uninstall

```
sudo dpkg -r engntru-debug
sudo dpkg -r openssl-111-debug # if installed from source
```

## Usage

### List algorithms

`openssl engine -c engntru`

### TLS test

#### Generate an RSA cert for the server

```
    openssl req -x509 -sha256 -nodes -newkey rsa:2048 -keyout /tmp/rsa.key -days 730 -out /tmp/rsa.cert -subj '/C=FI/ST=Uusimaa/L=Helsinki/CN=localhost' -config /etc/ssl/openssl.cnf

```

#### Launch the server

```
    openssl s_server -engine engntru -tls1_3 -key /tmp/rsa.key -cert /tmp/rsa.cert
```

#### Launch the client

```
    openssl s_client -engine engntru -tls1_3 -groups "SNTRUP761"
    # add -no_ticket -reconnect -async for multithread client testing
    # you might want to patch apps/s_client.c to increase the number of reconnections
```

    [`-groups`][SSL_CONF_cmd] takes a colon separated list of groups in order of
    preference, controlling, for clients, what is advertised as supported groups
    (and in TLS 1.3 also what is used in `key_share`) and, in servers, to
    determine which supported group to pick (first shared match).

[SSL_CONF_cmd]: https://www.openssl.org/docs/man1.1.1/man3/SSL_CONF_cmd.html

### ENV variables

#### `ENGNTRU_DEBUG`

The verbosity level of engNTRU logging output is controlled by the
`ENGNTRU_DEBUG` environment variable, which can be set to an integer value
as detailed in this list:

-   1 (LOG_FATAL)
-   2 (LOG_ERR)
-   3 (LOG_WARN) **default**
-   4 (LOG_INFO)
-   5 (LOG_DBG)
-   6 (LOG_VRB)
-   10 (LOG_EXTRM)

**NOTE**: Using a negative value will set the verbosity level to the
specified absolute value, but will activate _brief debug mode_,
shortening each line prefix.

_Example_: `export ENGNTRU_DEBUG=5` will setup engNTRU to print messages
with priority `LOG_DBG` or higher (highest priority is `LOG_FATAL`).

#### OpenSSL ENV variables

Environment variables relevant to the project include those affecting
OpenSSL, specifically the following ones:

-   `OPENSSL_ENGINES` sets the directory from which engines are loaded
    (the default value can be obtained by `openssl version -e`
-   `OPENSSL_CONF` sets a custom configuration file (the default value
    is `$OPENSSLDIR/openssl.cnf`, `openssl version -d`)

## License

engNTRU is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

engNTRU is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

The full text of the license is contained in the files `COPYING` and
`COPYING.LESSER`.

## Acknowledgments

-   Supported in part by Academy of Finland grant 303814.
-   This article is based in part upon work from COST Action IC1403
    CRYPTACUS, supported by COST (European Cooperation in Science and
    Technology).
