#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "crypto_kem.h"

struct kem_keypair_st {
    unsigned char pk[crypto_kem_PUBLICKEYBYTES];
    unsigned char sk[crypto_kem_SECRETKEYBYTES];
};

static inline
void print_hexstring(FILE *f, char *label, unsigned char buf[], size_t len)
{
    if (label) {
        fprintf(f, "%s", label);
    }
    for(size_t i = 0; i < len; i++) {
        fprintf(f, "%02X", buf[i]);
    }
    fprintf(f, "\n");
}

enum status_code {
    OK=0,
    ERROR=1,
};

static
int test_enc_dec(const struct kem_keypair_st *kp, int verbose)
{
    unsigned char ss[crypto_kem_BYTES]={0};
    unsigned char ct[crypto_kem_CIPHERTEXTBYTES]={0};
    unsigned char pt[crypto_kem_BYTES]={0};

    if (crypto_kem_enc(ct, ss, kp->pk) != 0) {
        fprintf(stderr, "ENC failed\n");
        return ERROR;
    }

    if (verbose) {
        print_hexstring(stdout, "Shared secret: ", ss, crypto_kem_BYTES);
        print_hexstring(stdout, "Ciphertext   : ", ct, crypto_kem_CIPHERTEXTBYTES);
    }

    if (crypto_kem_dec(pt, ct, kp->sk) != 0) {
        fprintf(stderr, "DEC failed\n");
        return ERROR;
    }

    if (verbose) {
        print_hexstring(stdout, "Plaintext    : ", pt, crypto_kem_BYTES);
    }

    if (memcmp(ss, pt, crypto_kem_BYTES) != 0) {
        fprintf(stderr, "CMP failed\n");
        return ERROR;
    }

    return OK;
}

static
int test_keygen(struct kem_keypair_st *kp)
{
    if (crypto_kem_keypair(kp->pk, kp->sk) != 0) {
        fprintf(stderr, "KEYGEN failed\n");
        return ERROR;
    }
    return OK;
}

static
int test_batch(unsigned n, int verbose)
{
    int ret = ERROR;
    unsigned char *sk_batch = NULL, *pk_batch = NULL;

    sk_batch = calloc(n, crypto_kem_SECRETKEYBYTES);
    pk_batch = calloc(n, crypto_kem_PUBLICKEYBYTES);

    if (sk_batch == NULL || pk_batch == NULL) {
        fprintf(stderr, "CALLOC failed\n");
        goto end;
    }

    if (crypto_kem_keypair_batch(pk_batch, sk_batch, n) != 0 ) {
        fprintf(stderr, "KEYPAIR_BATCH failed\n");
        goto end;
    }

    ret = 0;
    for (unsigned i = 0; i < n; i++) {
        struct kem_keypair_st kp;

        memcpy(kp.sk, sk_batch + i * crypto_kem_SECRETKEYBYTES, crypto_kem_SECRETKEYBYTES);
        memcpy(kp.pk, pk_batch + i * crypto_kem_PUBLICKEYBYTES, crypto_kem_PUBLICKEYBYTES);

        if (test_enc_dec(&kp, verbose) != OK) {
            ret++;
        }
    }

    if (ret > 0)
        ret = ERROR;
    else
        ret = OK;

end:
    if (sk_batch != NULL) free(sk_batch);
    if (pk_batch != NULL) free(pk_batch);

    return ret;
}



int main(void)
{
    struct kem_keypair_st kp;

    memset(&kp, 0, sizeof(kp));

    if (test_keygen(&kp) != OK
            || test_enc_dec(&kp, 1)) {
        return ERROR;
    }


    if (test_batch(crypto_kem_KEYBATCH, 0) != OK) {
        fprintf(stderr, "test_batch failed\n");
        return ERROR;
    }

    fprintf(stdout, "OK\n");
    return OK;
}
